@@ -12,6 +12,71 @@ from operator import attrgetter, itemgetter
 import operator
 import collections
 
+import Queue
+
+import logging
+import inspect
+
+from pickle import PicklingError
+
+import time
+logger = logging.getLogger('pyf.manager.network')
+
+class CustEllipsis(object):
+    pass
+
+BYPASS_VALS.append(CustEllipsis)
+
+class FlowEnd(object):
+    pass
+
+class FlowComplete(object):
+    pass
+
+class SerializedContainer(object):
+    def __init__(self, obj):
+        if hasattr(obj, 'serialized'):
+            self.serialized_obj = obj.serialized
+        elif hasattr(obj, 'serialize'):
+            self.serialized_obj = obj.serialize()
+        else:
+            raise ValueError, "object is not serializable"
+        
+        self.obj_class = obj.__class__
+    
+    def deserialize(self):
+        return self.obj_class.deserialize(self.serialized_obj)
+    
+    @classmethod
+    def is_serializable(cls, obj):
+        return (hasattr(obj, 'serialized') or hasattr(obj, 'serialize')) and hasattr(obj.__class__, "deserialize") 
+    
+def subprocess_output_adapter(flow, queue):
+    try:
+        for item in flow:
+            if item in BYPASS_VALS:
+                item = CustEllipsis
+                
+            elif SerializedContainer.is_serializable(item):
+                item = SerializedContainer(item)
+                
+            try:
+                queue.put(item)
+            except PicklingError:
+                raise
+                
+        queue.put(FlowComplete)
+        
+    except Exception, e:
+        if isinstance(e, StopIteration):
+            raise
+        else:
+            queue.put(e)
+            raise 
+        
+    finally:
+        queue.put(FlowEnd)
+
 class DataJunction(object):
     """ A DataJunction is an object specialised in taking multiple sources,
     setting up a joiner runner and permitting to get a generator output.
@@ -90,6 +155,46 @@ class DataJunction(object):
             return self.runner('out')
 
     __iter__ = get_generator
+    
+class GeneratorAdapter(object):
+    def __init__(self, queue):
+        self.queue = queue
+        
+    def __iter__(self):
+        import Queue
+        queue = self.queue
+        finished = False
+        complete = False
+        i = 0
+        while True:
+            if self.queue.empty() and i < 10:
+                if finished:
+                    if complete:
+                        raise StopIteration
+                    else:
+                        raise Exception, \
+                              'Bad ending in flow. Corrupted connection ?'
+                yield Ellipsis
+                i += 1
+            else:
+                i = 0
+                item = queue.get()
+                
+                if isinstance(item, Exception):
+                    raise item
+                
+                elif item is FlowEnd:
+                    finished = True
+                    
+                elif item is FlowComplete:
+                    complete = True 
+                
+                elif isinstance(item, SerializedContainer):
+                    yield item.deserialize()
+                    
+                else:
+                    yield item
+                    
 
 class Node(object):
     """ Object that is here to be linked to other nodes.
@@ -97,15 +202,17 @@ class Node(object):
     
     First step is to prelink, then to link.
     Order is from consumers up to producers :
-    - Consumers take output from up nodes than themselves take sources up...
-    - If there is multiple source, the joiner argument is used (or a default)
-    and encapsulated in a DataJunction
-    - If there is multiple output, the splitter argument is used (or a default)
+
+        * Consumers take output from up nodes than themselves take sources up...
+        * If there is multiple source, the joiner argument is used (or a default)
+          and encapsulated in a DataJunction
+        * If there is multiple output, the splitter argument is used (or a default)
     """
     def __init__(self, node_id, orunner,
                  in_port=None, out_port=None,
                  joiner=None, splitter=None,
-                 joiner_strip_bypass=True):
+                 joiner_strip_bypass=True,
+                 deffer_to_process=False):
         self._id = node_id
         self.input_ports = list()
         self.output_ports = list()
@@ -113,6 +220,16 @@ class Node(object):
         self.out_port_name = out_port or "out"
         self.joiner = joiner
         self.joiner_strip_bypass = joiner_strip_bypass
+        
+        
+        if deffer_to_process is True:
+            try:
+                import multiprocessing
+            except ImportError:
+                logger.warning('Asked to deffer to process, but multiprocessing is not available. Ignoring.')
+                deffer_to_process = False
+        
+        self.deffer_to_process = deffer_to_process
 
         self.junction = None
         self.splitter = splitter
@@ -155,6 +272,17 @@ class Node(object):
     def is_producer(self):
         """ Returns true if there are no input ports """
         return len(self.input_ports) == 0
+    
+    def prepare_splitter(self, splitter_function, size):
+        args = inspect.getargspec(splitter_function).args
+        kwargs = dict()
+        if "size" in args:
+            kwargs['size'] = size
+        if "ports" in args:
+            kwargs['ports'] = map(operator.attrgetter('name'),
+                                  self.output_ports)
+        
+        return runner(splitter_function, kwargs, handle_bypass=False)
         
     def get_default_splitter(self, size):
         """ Returns a default splitter (basically the same than
@@ -179,6 +307,7 @@ class Node(object):
         
         Prelinks the input nodes... (ask up to do the same)
         """
+        
         self.linked = True
         if not self.is_producer:
             for up_node in self.input_ports:
@@ -190,8 +319,12 @@ class Node(object):
                                              strip_bypass=self.joiner_strip_bypass)
         
         if len(self.output_ports) > 1:
-            self.splitter = self.splitter or\
-                            self.get_default_splitter(len(self.output_ports))
+            if self.splitter:
+                if inspect.isfunction(self.splitter):
+                    self.splitter = self.prepare_splitter(self.splitter,
+                                                        len(self.output_ports))
+            else:
+                self.splitter = self.get_default_splitter(len(self.output_ports))
                             
     def prepare_input_flow(self):
         while True:
@@ -203,9 +336,31 @@ class Node(object):
             if value in BYPASS_VALS:
                 self.bypass_buffer.append(value)
             else:
-                yield value
+                if isinstance(value, SerializedContainer):
+                    yield value.deserialize()
+                else:
+                    yield value
                 
     def prepare_output_flow(self, flow):
+#        output_queue = DataFlowQueue()
+        if self.deffer_to_process:
+            import multiprocessing
+            from multiprocessing.queues import SimpleQueue
+            output_queue = SimpleQueue()
+            
+            process = multiprocessing.Process(target=subprocess_output_adapter,
+                                              args=(self.__prepare_output_flow(flow),output_queue))
+            process.start()
+            
+            for itemm in iter(GeneratorAdapter(output_queue)):
+                yield itemm
+            
+#            process.join()
+        else:
+            for itemm in self.__prepare_output_flow(flow):
+                yield itemm
+        
+    def __prepare_output_flow(self, flow):
         while True:
             if not self.is_producer:
                 if self.input_buffer is None:
@@ -239,11 +394,11 @@ class Node(object):
         if not self.is_producer:
             if self.junction:
                 for up_node in self.input_ports:
-                    self.junction.add_input_source(up_node.get_output(),
+                    self.junction.add_input_source(up_node.get_output(self),
                                                    up_node.name)
                 self.input_flow = self.junction.get_generator()
             else:
-                self.input_flow = self.input_ports[0].get_output()
+                self.input_flow = self.input_ports[0].get_output(self)
                 
             self.runner.connect_in(self.in_port_name,
                                                  self.prepare_input_flow())
@@ -258,13 +413,13 @@ class Node(object):
         self.inited = True
             
         
-    def get_output(self):
+    def get_output(self, node=None):
         """ Returns an output generator.
         If there is one output, return the main runner out port,
         else, return a free port from the splitter """
         if self.splitter:
-            flow = self.splitter('out', self.current_output)
-            self.current_output += 1
+            flow = self.splitter('out', self.output_ports.index(node))
+#            self.current_output += 1
         else:
             flow = self.prepare_output_flow(self.runner(self.out_port_name))
             
@@ -291,7 +446,7 @@ class Node(object):
         
     def get_output_buffers(self):
         if not self.is_consumer:
-            if len(self.output_ports) > 1:
+            if len(self.output_ports) > 1 and 'out' in self.splitter.out_array_lookup:
                 return map(len, self.splitter.out_array_lookup['out'])
         
         return list()
@@ -299,18 +454,23 @@ class Node(object):
     def get_buffer_state(self, flow):
         position = self.get_output_port_position(flow)
         if not self.is_consumer and len(self.output_ports) > 1:
-            return len(self.splitter.out_array_lookup['out'][position])
+            if 'out' in self.splitter.out_array_lookup:
+                # initialized
+                return len(self.splitter.out_array_lookup['out'][position])
+            else:
+                # not initialized yet...
+                return 0
         else:
             return self.get_input_buffer()
             
 
 class Network(object):
-    """ Initialises a network of nodes from a dictionnary structure.
+    """ Initialises a network of nodes from a dictionary structure.
     Once the nodes have been set up, you can either get the consumer iterators,
     or get a status handler checking that all the outputs resolve to True.
-    """
-    def __init__(self, nodes=None):
-        """ 
+    
+    .. code-block::
+    
             [node1]
                |
             [node2]
@@ -318,13 +478,15 @@ class Network(object):
         [node3]  [node4]
              \   /
             [node5]
-            
+    
+    .. code-block:: python        
         nodes => {'node1': {component=my_producer, after=['node2']},
                   'node2': {component=tata, after=['node4']},
                   'node3': {component=titi, after=['node5'], before=['node2']},
                   'node4': {component=tutu, after=['node5']},
                   'node5': {component=my_consumer}}
-        """
+    """
+    def __init__(self, nodes=None):
         self.ids = list()
         self.nodes = nodes or dict()
         self.node_objects = dict()
@@ -364,7 +526,8 @@ class Network(object):
                                              out_port=node_data.get('out_port'),
                                              joiner=node_data.get('joiner'),
                                              splitter=node_data.get('splitter'),
-                                             joiner_strip_bypass=node_data.get('joiner_strip_bypass', True))
+                                             joiner_strip_bypass=node_data.get('joiner_strip_bypass', True),
+                                             deffer_to_process=node_data.get('deffer_to_process', False))
         
         for node_id, node_data in self.nodes.iteritems():
             node = self.node_objects[node_id]
